/*
  spamdyke -- a filter for stopping spam at connection time.
  Copyright (C) 2007 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <syslog.h>
#include <signal.h>
#include <sys/select.h>
#include <time.h>
#include <stdlib.h>
#include "spamdyke.h"
#include "environment.h"
#include "log.h"
#include "configuration.h"
#include "exec.h"

int exec_path(char *filename, char *argv[], char *envp[])
  {
  char *path;
  char *tmp_start;
  char *tmp_end;
  char new_filename[MAX_BUF + 1];

  execve(filename, argv, envp);

  if (strchr(filename, '/') == NULL)
    {
    if (((path = find_environment_variable(envp, ENVIRONMENT_PATH, STRLEN(ENVIRONMENT_PATH))) == NULL) ||
        (path[0] == '\0'))
      path = DEFAULT_PATH;

    tmp_start = path;
    tmp_end = NULL;
    while (tmp_start != NULL)
      {
      if ((tmp_end = strchr(tmp_start, ':')) != NULL)
        {
        snprintf(new_filename, MAX_BUF, "%.*s/%s", (int)(tmp_end - tmp_start), tmp_start, filename);
        tmp_start = tmp_end + 1;
        }
      else
        {
        snprintf(new_filename, MAX_BUF, "%s/%s", tmp_start, filename);
        tmp_start = NULL;
        }

      execve(new_filename, argv, envp);
      }
    }

  return(0);
  }

int exec_checkpassword(struct filter_settings *current_settings, char *command_line, char *username, char *password, char *timestamp, char *envp[])
  {
  int return_value;
  int i;
  char **child_argv;
  int stdout_pipe[2];
  int stderr_pipe[2];
  int output_pipe[2];
  pid_t child_pid;
  pid_t last_wait;
  char output_buf[MAX_CHECKPASSWORD + 1];
  int strlen_output_buf;
  int strlen_sent_buf;
  int status;
  char *last_arg;
  char *tmp_arg;
  fd_set read_fds;
  fd_set write_fds;
  int max_fd;
  struct timeval timeout;
  time_t start_time;
  char tmp_char;

  return_value = 0;
  child_argv = NULL;
  last_wait = 0;

  if (pipe(output_pipe) != -1)
    if (pipe(stdout_pipe) != -1)
      if (pipe(stderr_pipe) != -1)
        if ((child_pid = fork()) > 0)
          {
          close(output_pipe[0]);
          close(stdout_pipe[1]);
          close(stderr_pipe[1]);

          strlen_output_buf = snprintf(output_buf, MAX_CHECKPASSWORD, "%s%c%s%c%s%c", (username != NULL) ? username : "", '\0', (password != NULL) ? password : "", '\0', (timestamp != NULL) ? timestamp : "", '\0');
          strlen_sent_buf = 0;

          start_time = time(NULL);

          timeout.tv_sec = TIMEOUT_CHECKPASSWORD_SECS - (time(NULL) - start_time);
          timeout.tv_usec = 0;

          FD_ZERO(&read_fds);
          FD_ZERO(&write_fds);
          FD_SET(output_pipe[1], &write_fds);
          FD_SET(stdout_pipe[0], &read_fds);
          FD_SET(stderr_pipe[0], &read_fds);
          max_fd = MAXVAL(MAXVAL(output_pipe[1], stdout_pipe[0]), stderr_pipe[0]);

          while (select(max_fd + 1, &read_fds, &write_fds, NULL, &timeout) >= 0)
            {
            if (FD_ISSET(stdout_pipe[0], &read_fds) &&
                (read(stdout_pipe[0], &tmp_char, 1) == 0))
              break;

            if (FD_ISSET(stderr_pipe[0], &read_fds) &&
                (read(stderr_pipe[0], &tmp_char, 1) == 0))
              break;

            if ((output_pipe[1] >= 0) &&
                FD_ISSET(output_pipe[1], &write_fds))
              {
              write(output_pipe[1], output_buf + strlen_sent_buf, 1);
              strlen_sent_buf++;
              }

            timeout.tv_sec = TIMEOUT_CHECKPASSWORD_SECS - (time(NULL) - start_time);
            timeout.tv_usec = 0;

            if (((last_wait = waitpid(child_pid, &status, WNOHANG)) == 0) &&
                (timeout.tv_sec > 0))
              {
              FD_ZERO(&read_fds);
              FD_ZERO(&write_fds);
              FD_SET(stdout_pipe[0], &read_fds);
              FD_SET(stderr_pipe[0], &read_fds);

              if (strlen_sent_buf < strlen_output_buf)
                {
                FD_SET(output_pipe[1], &write_fds);
                max_fd = MAXVAL(MAXVAL(output_pipe[1], stdout_pipe[0]), stderr_pipe[0]);
                }
              else
                {
                if (output_pipe[1] != -1)
                  {
                  close(output_pipe[1]);
                  output_pipe[1] = -1;
                  }

                timeout.tv_sec = 0;
                timeout.tv_usec = 500;
                max_fd = MAXVAL(stdout_pipe[0], stderr_pipe[0]);
                }
              }
            else
              break;
            }

          if (last_wait == 0)
            waitpid(child_pid, &status, 0);

          if (WIFEXITED(status))
            {
            /* Status codes are documented at http://cr.yp.to/checkpwd/interface.html */
            switch (WEXITSTATUS(status))
              {
              case 0:
                /* Success */
                spamdyke_log(current_settings, LOG_LEVEL_DEBUG, LOG_ERROR_AUTH_SUCCESS "%s", username);
                return_value = 1;
                break;
              case 1:
                /* Bad username/password */
                spamdyke_log(current_settings, LOG_LEVEL_INFO, LOG_ERROR_AUTH_FAILURE "%s", username);
                break;
              case 2:
                /* Misuse?  WTF is "misuse"? */
                spamdyke_log(current_settings, LOG_LEVEL_INFO, LOG_ERROR_AUTH_MISUSE "%s", username);
                break;
              case 111:
                /* Temporary error */
                spamdyke_log(current_settings, LOG_LEVEL_INFO, LOG_ERROR_AUTH_ERROR "%s", username);
                break;
              default:
                /* Undocumented error code */
                spamdyke_log(current_settings, LOG_LEVEL_INFO, LOG_ERROR_AUTH_UNKNOWN "(%d): %s", WEXITSTATUS(status), username);
                break;
              }
            }
          else
            {
            kill(child_pid, SIGKILL);
            spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_AUTH_ABEND "%s", username);
            }

          if (output_pipe[1] != -1)
            close(output_pipe[1]);
          close(stdout_pipe[0]);
          close(stderr_pipe[0]);
          }
        else if (child_pid == 0)
          {
          close(output_pipe[1]);
          close(stdout_pipe[0]);
          close(stderr_pipe[0]);

          if (dup2(output_pipe[0], CHECKPASSWORD_FD) != -1)
            if (dup2(stdout_pipe[1], STDOUT_FD) != -1)
              if (dup2(stderr_pipe[1], STDERR_FD) != -1)
                {
                last_arg = command_line;
                while ((tmp_arg = strchr(last_arg, ' ')) != NULL)
                  {
                  append_string(current_settings, &child_argv, last_arg, tmp_arg - last_arg);
                  last_arg = tmp_arg + 1;
                  }
                append_string(current_settings, &child_argv, last_arg, strlen(last_arg));

                if (!exec_path(child_argv[0], child_argv, envp))
                  {
                  spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_EXEC "%s: %s", child_argv[0], strerror(errno));

                  close(output_pipe[0]);
                  close(stdout_pipe[1]);
                  close(stderr_pipe[1]);

                  for (i = 0; child_argv[i] != NULL; i++)
                    free(child_argv[i]);
                  free(child_argv);

                  exit(111);
                  }
                }
              else
                {
                spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_MOVE_DESCRIPTORS "%s", strerror(errno));

                close(output_pipe[0]);
                close(stdout_pipe[1]);
                close(stderr_pipe[1]);
                }
            else
              {
              spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_MOVE_DESCRIPTORS "%s", strerror(errno));

              close(output_pipe[0]);
              close(stdout_pipe[1]);
              close(stderr_pipe[1]);
              }
          else
            {
            spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_MOVE_DESCRIPTORS "%s", strerror(errno));

            close(output_pipe[0]);
            close(stdout_pipe[1]);
            close(stderr_pipe[1]);
            }
          }
        else
          {
          spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_FORK "%s", strerror(errno));

          close(stderr_pipe[0]);
          close(stderr_pipe[1]);
          close(stdout_pipe[0]);
          close(stdout_pipe[1]);
          close(output_pipe[0]);
          close(output_pipe[1]);
          }
      else
        {
        spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_PIPE "%s", strerror(errno));

        close(stdout_pipe[0]);
        close(stdout_pipe[1]);
        close(output_pipe[0]);
        close(output_pipe[1]);
        }
    else
      {
      spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_PIPE "%s", strerror(errno));

      close(output_pipe[0]);
      close(output_pipe[1]);
      }
  else
    spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_PIPE "%s", strerror(errno));

  return(return_value);
  }

int exec_command(struct filter_settings *current_settings, char *command_line, char *envp[], char **return_content, int size_return_content)
  {
  int return_value;
  int i;
  char **child_argv;
  int stdout_pipe[2];
  int stderr_pipe[2];
  int output_pipe[2];
  pid_t child_pid;
  pid_t last_wait;
  char input_buf[MAX_COMMAND_BUF + 1];
  int strlen_input_buf;
  int status;
  char *last_arg;
  char *tmp_arg;
  fd_set read_fds;
  int max_fd;
  struct timeval timeout;
  time_t start_time;
  char tmp_char;
  int strlen_input_total;
  char *tmp_alloc;
  char *tmp_return_content;
  int tmp_strlen;

  return_value = -1;
  child_argv = NULL;
  last_wait = 0;
  tmp_return_content = NULL;

  if ((command_line != NULL) &&
      (return_content != NULL))
    {
    if (pipe(output_pipe) != -1)
      if (pipe(stdout_pipe) != -1)
        if (pipe(stderr_pipe) != -1)
          if ((child_pid = fork()) > 0)
            {
            close(output_pipe[0]);
            close(stdout_pipe[1]);
            close(stderr_pipe[1]);

            start_time = time(NULL);

            timeout.tv_sec = TIMEOUT_COMMAND_SECS - (time(NULL) - start_time);
            timeout.tv_usec = 0;

            FD_ZERO(&read_fds);
            FD_SET(stdout_pipe[0], &read_fds);
            FD_SET(stderr_pipe[0], &read_fds);
            max_fd = MAXVAL(stdout_pipe[0], stderr_pipe[0]);

            return_value = 0;
            strlen_input_buf = 0;
            strlen_input_total = 0;

            while (select(max_fd + 1, &read_fds, NULL, NULL, &timeout) >= 0)
              {
              if ((FD_ISSET(stdout_pipe[0], &read_fds) &&
                   (read(stdout_pipe[0], &tmp_char, 1) == 0)) ||
                  (FD_ISSET(stderr_pipe[0], &read_fds) &&
                   (read(stderr_pipe[0], &tmp_char, 1) == 0)))
                {
                input_buf[strlen_input_buf] = tmp_char;
                strlen_input_buf++;
                strlen_input_total++;

                if (strlen_input_buf == MAX_COMMAND_BUF)
                  {
                  if (size_return_content == -1)
                    if ((tmp_alloc = realloc(tmp_return_content, strlen_input_total + 1)) != NULL)
                      {
                      strncpy(tmp_alloc + (strlen_input_total - strlen_input_buf), input_buf, strlen_input_buf);
                      tmp_alloc[strlen_input_total] = '\0';
                      tmp_return_content = tmp_alloc;
                      }
                    else
                      {
                      spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_REALLOC, strlen_input_total + 1);
                      return_value = -1;
                      break;
                      }
                  else if ((strlen_input_total - strlen_input_buf) < size_return_content)
                    {
                    tmp_strlen = MINVAL(size_return_content, strlen_input_total);
                    strncpy(*return_content, input_buf, tmp_strlen);
                    (*return_content)[tmp_strlen] = '\0';
                    }

                  strlen_input_buf = 0;
                  }
                }

              timeout.tv_sec = TIMEOUT_CHECKPASSWORD_SECS - (time(NULL) - start_time);
              timeout.tv_usec = 0;

              if (((last_wait = waitpid(child_pid, &status, WNOHANG)) == 0) &&
                  (timeout.tv_sec > 0))
                {
                FD_ZERO(&read_fds);
                FD_SET(stdout_pipe[0], &read_fds);
                FD_SET(stderr_pipe[0], &read_fds);
                }
              else
                break;
              }

            if (last_wait == 0)
              waitpid(child_pid, &status, 0);

            if (WIFEXITED(status))
              {
              if (size_return_content == -1)
                {
                if (*return_content != NULL)
                  free(*return_content);

                *return_content = tmp_return_content;
                tmp_return_content = NULL;
                return_value = strlen_input_total;
                }
              else
                return_value = MINVAL(size_return_content, strlen_input_total);
              }
            else
              {
              kill(child_pid, SIGKILL);
              spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_COMMAND_ABEND "%s", command_line);
              }

            close(output_pipe[1]);
            close(stdout_pipe[0]);
            close(stderr_pipe[0]);
            }
          else if (child_pid == 0)
            {
            close(output_pipe[1]);
            close(stdout_pipe[0]);
            close(stderr_pipe[0]);

            if (dup2(output_pipe[0], CHECKPASSWORD_FD) != -1)
              if (dup2(stdout_pipe[1], STDOUT_FD) != -1)
                if (dup2(stderr_pipe[1], STDERR_FD) != -1)
                  {
                  last_arg = command_line;
                  while ((tmp_arg = strchr(last_arg, ' ')) != NULL)
                    {
                    append_string(current_settings, &child_argv, last_arg, tmp_arg - last_arg);
                    last_arg = tmp_arg + 1;
                    }
                  append_string(current_settings, &child_argv, last_arg, strlen(last_arg));

                  if (!exec_path(child_argv[0], child_argv, envp))
                    {
                    spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_EXEC "%s: %s", child_argv[0], strerror(errno));

                    close(output_pipe[0]);
                    close(stdout_pipe[1]);
                    close(stderr_pipe[1]);

                    for (i = 0; child_argv[i] != NULL; i++)
                      free(child_argv[i]);
                    free(child_argv);

                    exit(0);
                    }
                  }
                else
                  {
                  spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_MOVE_DESCRIPTORS "%s", strerror(errno));

                  close(output_pipe[0]);
                  close(stdout_pipe[1]);
                  close(stderr_pipe[1]);
                  }
              else
                {
                spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_MOVE_DESCRIPTORS "%s", strerror(errno));

                close(output_pipe[0]);
                close(stdout_pipe[1]);
                close(stderr_pipe[1]);
                }
            else
              {
              spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_MOVE_DESCRIPTORS "%s", strerror(errno));

              close(output_pipe[0]);
              close(stdout_pipe[1]);
              close(stderr_pipe[1]);
              }
            }
          else
            {
            spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_FORK "%s", strerror(errno));

            close(stderr_pipe[0]);
            close(stderr_pipe[1]);
            close(stdout_pipe[0]);
            close(stdout_pipe[1]);
            close(output_pipe[0]);
            close(output_pipe[1]);
            }
        else
          {
          spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_PIPE "%s", strerror(errno));

          close(stdout_pipe[0]);
          close(stdout_pipe[1]);
          close(output_pipe[0]);
          close(output_pipe[1]);
          }
      else
        {
        spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_PIPE "%s", strerror(errno));

        close(output_pipe[0]);
        close(output_pipe[1]);
        }
    else
      spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_PIPE "%s", strerror(errno));
    }

  if (tmp_return_content != NULL)
    free(tmp_return_content);

  return(return_value);
  }
