/*
  passwordcheck -- a simple program for testing checkpassword
  Copyright (C) 2007 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <getopt.h>
#include <errno.h>
#include "md5.h"
#include "config.h"

#define MAX_BUF                         4095
#define MAX_CHECKPASSWORD               511
#define TIMEOUT_CHECKPASSWORD_SECS      30
#define ENVIRONMENT_PATH                "PATH"
#define DEFAULT_PATH                    "/bin:/usr/bin:/usr/local/bin:/var/qmail/bin:/home/vpopmail/bin"
#define ENVIRONMENT_DELIMITER           ':'

#define IPAD_BYTE                       0x36
#define OPAD_BYTE                       0x5C

#define STDIN_FD                        0
#define STDOUT_FD                       1
#define STDERR_FD                       2
#define CHECKPASSWORD_FD                3

#define MAXVAL(a,b)                     (((a) < (b)) ? (b) : (a))
#define STRLEN(X)                       (sizeof(X) - 1)

FILE *stderr_file = NULL;

int append_string(char ***target_list, char *target_string, int strlen_target_string)
  {
  int return_value;
  int i;
  char *tmp_char;
  char **tmp_array;

  return_value = 0;

  if ((*target_list) != NULL)
    for (i = 0; (*target_list)[i] != NULL; i++);
  else
    i = 0;

  tmp_char = NULL;
  if ((strlen_target_string == -1) ||
      ((tmp_char = malloc(strlen_target_string + 1)) != NULL))
    if ((tmp_array = realloc((*target_list), sizeof(char *) * (i + 2))) != NULL)
      {
      if (strlen_target_string >= 0)
        {
        memcpy(tmp_char, target_string, strlen_target_string);
        tmp_char[strlen_target_string] = '\0';
        tmp_array[i] = tmp_char;
        }
      else
        tmp_array[i] = target_string;

      tmp_array[i + 1] = NULL;
      *target_list = tmp_array;

      return_value = 1;
      }
    else
      {
      printf("ERROR: Unable to allocate %ld bytes\n", (unsigned long)(sizeof(char *) * (i + 2)));
      free(tmp_char);
      }
  else
    printf("ERROR: Unable to allocate %ld bytes\n", (unsigned long)(sizeof(char *) * (i + 2)));

  return(return_value);
  }

char *find_environment_variable(char **environment, char *target_variable, int strlen_target_variable)
  {
  char *return_value;
  int i;

  return_value = NULL;

  if (environment != NULL)
    for (i = 0; environment[i] != NULL; i++)
      if ((strncmp(environment[i], target_variable, strlen_target_variable) == 0) &&
          (environment[i][strlen_target_variable] == ENVIRONMENT_DELIMITER))
        {
        return_value = environment[i] + strlen_target_variable + 1;
        break;
        }

  return(return_value);
  }

int exec_path(char *filename, char *argv[], char *envp[])
  {
  char *path;
  char *tmp_start;
  char *tmp_end;
  char new_filename[MAX_BUF + 1];

  execve(filename, argv, envp);

  if (strchr(filename, '/') == NULL)
    {
    if (((path = find_environment_variable(envp, ENVIRONMENT_PATH, STRLEN(ENVIRONMENT_PATH))) == NULL) ||
        (path[0] == '\0'))
      path = DEFAULT_PATH;

    tmp_start = path;
    tmp_end = NULL;
    while (tmp_start != NULL)
      {
      if ((tmp_end = strchr(tmp_start, ':')) != NULL)
        {
        snprintf(new_filename, MAX_BUF, "%.*s/%s", (int)(tmp_end - tmp_start), tmp_start, filename);
        tmp_start = tmp_end + 1;
        }
      else
        {
        snprintf(new_filename, MAX_BUF, "%s/%s", tmp_start, filename);
        tmp_start = NULL;
        }

      execve(new_filename, argv, envp);
      }
    }

  return(0);
  }

int exec_checkpassword(char **child_argv, char *username, char *password, char *timestamp, char *envp[])
  {
  int return_value;
  //int i;
  //char **child_argv;
  int stdout_pipe[2];
  //int stderr_pipe[2];
  int output_pipe[2];
  pid_t child_pid;
  pid_t last_wait;
  char output_buf[MAX_CHECKPASSWORD + 1];
  int strlen_output_buf;
  int strlen_sent_buf;
  int status;
  //char *last_arg;
  //char *tmp_arg;
  fd_set read_fds;
  fd_set write_fds;
  int max_fd;
  struct timeval timeout;
  time_t start_time;
  char tmp_char;

  return_value = 0;
  //child_argv = NULL;
  last_wait = 0;

  if (pipe(output_pipe) != -1)
    if (pipe(stdout_pipe) != -1)
      //if (pipe(stderr_pipe) != -1)
        if ((child_pid = fork()) > 0)
          {
          close(output_pipe[0]);
          close(stdout_pipe[1]);
          //close(stderr_pipe[1]);

          printf("Sending data to child process: %s\\0%s\\0%s\\0\n", (username != NULL) ? username : "", (password != NULL) ? password : "", (timestamp != NULL) ? timestamp : "");
          strlen_output_buf = snprintf(output_buf, MAX_CHECKPASSWORD, "%s%c%s%c%s%c", (username != NULL) ? username : "", '\0', (password != NULL) ? password : "", '\0', (timestamp != NULL) ? timestamp : "", '\0');
          strlen_sent_buf = 0;

          start_time = time(NULL);

          timeout.tv_sec = TIMEOUT_CHECKPASSWORD_SECS - (time(NULL) - start_time);
          timeout.tv_usec = 0;

          FD_ZERO(&read_fds);
          FD_ZERO(&write_fds);
          FD_SET(output_pipe[1], &write_fds);
          FD_SET(stdout_pipe[0], &read_fds);
          //FD_SET(stderr_pipe[0], &read_fds);
          //max_fd = MAXVAL(MAXVAL(output_pipe[1], stdout_pipe[0]), stderr_pipe[0]);
          max_fd = MAXVAL(output_pipe[1], stdout_pipe[0]);

          while (select(max_fd + 1, &read_fds, &write_fds, NULL, &timeout) >= 0)
            {
            if (FD_ISSET(stdout_pipe[0], &read_fds) &&
                (read(stdout_pipe[0], &tmp_char, 1) == 0))
              break;

            /*
            if (FD_ISSET(stderr_pipe[0], &read_fds) &&
                (read(stderr_pipe[0], &tmp_char, 1) == 0))
              break;
            */

            if ((output_pipe[1] >= 0) &&
                FD_ISSET(output_pipe[1], &write_fds))
              {
              write(output_pipe[1], output_buf + strlen_sent_buf, 1);
              strlen_sent_buf++;
              }

            timeout.tv_sec = TIMEOUT_CHECKPASSWORD_SECS - (time(NULL) - start_time);
            timeout.tv_usec = 0;

            if (((last_wait = waitpid(child_pid, &status, WNOHANG)) == 0) &&
                (timeout.tv_sec > 0))
              {
              FD_ZERO(&read_fds);
              FD_ZERO(&write_fds);
              FD_SET(stdout_pipe[0], &read_fds);
              //FD_SET(stderr_pipe[0], &read_fds);

              if (strlen_sent_buf < strlen_output_buf)
                {
                FD_SET(output_pipe[1], &write_fds);
                //max_fd = MAXVAL(MAXVAL(output_pipe[1], stdout_pipe[0]), stderr_pipe[0]);
                max_fd = MAXVAL(output_pipe[1], stdout_pipe[0]);
                }
              else
                {
                if (output_pipe[1] != -1)
                  {
                  close(output_pipe[1]);
                  output_pipe[1] = -1;
                  }

                timeout.tv_sec = 0;
                timeout.tv_usec = 500;
                //max_fd = MAXVAL(stdout_pipe[0], stderr_pipe[0]);
                max_fd = stdout_pipe[0];
                }
              }
            else
              break;
            }

          if (last_wait == 0)
            waitpid(child_pid, &status, 0);

          if (WIFEXITED(status))
            {
            switch (WEXITSTATUS(status))
              {
              case 0:
                return_value = 1;
                break;
              case 1:
                printf("ERROR: Child process indicates incorrect password.\n");
                break;
              case 2:
                printf("ERROR: Child process indicates possible misuse.  Not sure what this means.\n");
                break;
              case 111:
                printf("ERROR: Child process indicates temporary error.  Check that %s exists and is executable.\n", child_argv[1]);
                break;
              default:
                printf("ERROR: Child process exited with an undocumented return code: %d\n", WEXITSTATUS(status));
              }
            }
          else
            {
            printf("ERROR: Child process did not exit normally.  Check that %s exists and is executable.\n", child_argv[0]);
            kill(child_pid, SIGKILL);
            }

          if (output_pipe[1] != -1)
            close(output_pipe[1]);
          close(stdout_pipe[0]);
          //close(stderr_pipe[0]);
          }
        else if (child_pid == 0)
          {
          close(output_pipe[1]);
          close(stdout_pipe[0]);
          //close(stderr_pipe[0]);

          stderr_file = fdopen(STDERR_FD, "w");

          if (dup2(output_pipe[0], CHECKPASSWORD_FD) != -1)
            if (dup2(stdout_pipe[1], STDOUT_FD) != -1)
              //if (dup2(stderr_pipe[1], STDERR_FD) != -1)
                {
                /*
                last_arg = command_line;
                while ((tmp_arg = strchr(last_arg, ' ')) != NULL)
                  {
                  append_string(&child_argv, last_arg, tmp_arg - last_arg);
                  last_arg = tmp_arg + 1;
                  }
                append_string(&child_argv, last_arg, strlen(last_arg));
                */

                if (!exec_path(child_argv[0], child_argv, envp))
                  {
                  fprintf(stderr_file, "ERROR: Unable to execute %s: %s\n", child_argv[0], strerror(errno));

                  close(output_pipe[0]);
                  close(stdout_pipe[1]);
                  //close(stderr_pipe[1]);

                  /*
                  for (i = 0; child_argv[i] != NULL; i++)
                    free(child_argv[i]);
                  free(child_argv);
                  */

                  exit(111);
                  }
                }
              /*
              else
                {
                fprintf(stderr_file, "ERROR: Unable to move descriptors with dup2(): %s\n", strerror(errno));

                close(output_pipe[0]);
                close(stdout_pipe[1]);
                close(stderr_pipe[1]);
                }
              */
            else
              {
              fprintf(stderr_file, "ERROR: Unable to move descriptors with dup2(): %s\n", strerror(errno));

              close(output_pipe[0]);
              close(stdout_pipe[1]);
              //close(stderr_pipe[1]);
              }
          else
            {
            fprintf(stderr_file, "ERROR: Unable to move descriptors with dup2(): %s\n", strerror(errno));

            close(output_pipe[0]);
            close(stdout_pipe[1]);
            //close(stderr_pipe[1]);
            }
          }
        else
          {
          printf("ERROR: Unable to spawn child with fork(): %s\n", strerror(errno));

          //close(stderr_pipe[0]);
          //close(stderr_pipe[1]);
          close(stdout_pipe[0]);
          close(stdout_pipe[1]);
          close(output_pipe[0]);
          close(output_pipe[1]);
          }
      /*
      else
        {
        printf("ERROR: Unable to create pipe with pipe(): %s\n", strerror(errno));

        close(stdout_pipe[0]);
        close(stdout_pipe[1]);
        close(output_pipe[0]);
        close(output_pipe[1]);
        }
      */
    else
      {
      printf("ERROR: Unable to create pipe with pipe(): %s\n", strerror(errno));

      close(output_pipe[0]);
      close(output_pipe[1]);
      }
  else
    printf("ERROR: Unable to create pipe with pipe(): %s\n", strerror(errno));

  return(return_value);
  }

void usage()
  {
  printf(
    PACKAGE_NAME " " PACKAGE_VERSION " (C)2007 Sam Clippinger, " PACKAGE_BUGREPORT "\n"
    "http://www.spamdyke.org/\n"
    "\n"
    "USAGE: passwordcheck -u USERNAME -p PASSWORD [ -- ] COMMANDLINE [ ARG1 ARG2 ... ]\n"
    "\n"
    "Runs the command COMMANDLINE and passes USERNAME and PASSWORD the same way\n"
    "spamdyke will do if COMMANDLINE is used as an SMTP AUTH command.\n"
    );

  exit(0);

  return;
  }

int main(int argc, char *argv[], char *envp[])
  {
  int i;
  int opt;
  int strlen_challenge;
  char *challenge;
  char *username;
  char *password;
  char **new_envp;
  int result;
  unsigned char ipad[MAX_BUF + 1];
  unsigned char opad[MAX_BUF + 1];
  unsigned char md5_result[16];
  unsigned char secret[64];
  unsigned char final[33];

  username = NULL;
  password = NULL;

  while ((opt = getopt(argc, argv, "u:p:")) != -1)
    {
    switch (opt)
      {
      case 'u':
        username = optarg;
        break;
      case 'p':
        password = optarg;
        break;
      default:
        usage();
      }
    }

  if ((username != NULL) &&
      (password != NULL) &&
      (argv[optind] != NULL))
    {
    for (i = 0; envp[i] != NULL; i++);
    if ((new_envp = (char **)malloc(sizeof(char *) * (i + 2))) != NULL)
      memcpy(new_envp, envp, sizeof(char *) * (i + 1));

    append_string(&new_envp, "TCPLOCALPORT=25", -1);
    printf("Beginning unencrypted password test.\n");
    result = exec_checkpassword(argv + optind, username, password, NULL, new_envp);
    printf("Test result: %s\n\n", (result == 1) ? "SUCCESS" : "FAILURE");

    if (result)
      {
      for (i = 0; i < 64; i++)
        {
        ipad[i] = IPAD_BYTE;
        opad[i] = OPAD_BYTE;
        }

      if (strlen(password) > 64)
        {
        md5(secret, (unsigned char *)password, strlen(password));
        for (i = 16; i < 64; i++)
          secret[i] = '\0';
        }
      else
        {
        strncpy((char *)secret, password, strlen(password));
        for (i = strlen(argv[2]); i < 64; i++)
          secret[i] = '\0';
        }

      for (i = 0; i < 64; i++)
        {
        ipad[i] ^= secret[i];
        opad[i] ^= secret[i];
        }

      challenge = "server.name.some.domain";
      strlen_challenge = strlen(challenge);
      for (i = 0; i < strlen_challenge; i++)
        ipad[i + 64] = challenge[i];

      md5(opad + 64, ipad, strlen_challenge + 64);
      md5(md5_result, opad, 80);

      for (i = 0; i < 16; i++)
        snprintf((char *)(final + (i * 2)), 33 - (i * 2), "%.2x", md5_result[i]);

      printf("Beginning CRAM-MD5 encrypted password test.\n");
      if ((result = exec_checkpassword(argv + optind, username, (char *)final, challenge, new_envp)) == 1)
        printf("Test result: SUCCESS\n");
      else
        printf("Test result: FAILURE (perhaps %s doesn't support CRAM-MD5)\n", argv[optind]);
      }
    }
  else
    usage();

  return(0);
  }
